"""
Helpers for array operations
"""
from typing import TypeVar, NewVar

import numpy as np

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'WARNING')


from aopp_deconv_tool.typedef import NumVar, ShapeVar

T = TypeVar('T')

type N = NewType('N', NumVar)
type M = NewType('M', NumVar)

type S[X] = NewType('ShapeS', ShapeVar[X])
type Q[X] = NewType('ShapeQ', ShapeVar[X])
type R[X] = NewType('ShapeR', ShapeVar[X])


def pacman(a : np.ndarray[S[N],T], delta : np.ndarray[[N],int]) -> np.ndarray[S[N],T]:
	"""
	Shift an array by some delta using pacman (periodic) boundary conditions.
	E.g. tiles of the same array exist to either side.
	"""
	return np.roll(a, delta, tuple(i for i in range(a.ndim)))

def periodic(a : np.ndarray[S[N],T], delta : np.ndarray[[N],int]) -> np.ndarray[S[N],T]:
	"""
	Shift an array by some delta using periodic (pacman) boundary conditions
	E.g. tiles of the same array exist to either side.
	"""
	return pacman(a,delta)

def const(a : np.ndarray[S[N],T], delta : np.ndarray[[N],int], const : T = 0) -> np.ndarray[S[N],T]:
	"""
	Shift and array by some delta using constant boundary conditions
	E.g. the array is surrounded by a constant value
	"""
	# Shift as in periodic boundary conditions, then fill the "empty" region
	# with a constant value
	s = pacman(a, delta)
	for i, d in enumerate(delta):
		wrap_region_t = tuple((slice(d,None) if d < 0 else slice(None,d)) if i==j else slice(None) for j, d in enumerate(delta))
		s[wrap_region_t] = const
	return s

def reflect(a : np.ndarray[S[N],T], delta : np.ndarray[[N],int]) -> np.ndarray[S[N],T]:
	"""
	Shift an array by some delta using reflecting boundary conditions.
	E.g. the array is surrounded by reflected versions of itself.
	"""
	# Shift using pacman as normal, then set the 'wrapped' region to the data 
	# for reflected boundary conditions
	s = pacman(a, delta)
	for i, d in enumerate(delta):
		n = a.shape[i]
		wrap_region_t = tuple((slice(d,n,1) if d < 0 else slice(0,d,1)) if i==j else slice(None) for j, d in enumerate(delta))
		reflect_region_t = tuple((slice(n-2,n+d-2,-1) if d < 0 else slice(d,0,-1)) if i==j else slice(None) for j, d in enumerate(delta))
		s[wrap_region_t] = a[reflect_region_t]
	return s
 
 



"""
Helper functions for slice operations on numpy arrays
"""
from typing import Iterator
import numpy as np

import aopp_deconv_tool.numpy_helper as nph
import aopp_deconv_tool.numpy_helper.axes

import aopp_deconv_tool.cast as cast

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'INFO')


def around_center(big_shape : tuple[int,...], small_shape : tuple[int,...]) -> tuple[slice,...]:
	"""
	returns a slice of a in the shape of b about the center of a
	Accepts np.ndarray or tuple for each argument
	"""
	s_diff = tuple(s1-s2 for s1,s2 in zip(big_shape, small_shape))
	return(tuple([slice(d//2, s-(d//2+d%2)) for s, d in zip(big_shape,s_diff)]))

def from_string(slice_tuple_str : str) -> tuple[slice,...]:
	"""
	Build a tuple of slices from a string representation
	"""
	try:
		return tuple(slice(*tuple(cast.to(z,int) if z != '' else None for z in y.split(':'))) for y in slice_tuple_str.split(','))
	except Exception as e:
		e.add_note(f"slice tuple string '{slice_tuple_str}' is malformed")
		raise

def squeeze(slice_tuple: tuple[slice | int,...]) -> tuple[slice | int]:
	"""
	Given a tuple of slices (`slice_tuple`), if a slice only selects a single index, replace it with an index instead.
	e.g. [10:100, 13:14, 50:60] -> [10:100, 13, 50:60]
	"""
	return tuple(s.start if type(s) is slice and ((s.stop - s.start) == (1 if s.step is None else s.step)) else s for s in slice_tuple )

def unsqueeze(slice_tuple: tuple[slice | int,...]) -> tuple[slice]:
	"""
	Given a tuple of slices and indices (`slice_tuple`), replace all indices with a slice from i->i+1
	e.g. [10:100, 13, 50:60] -> [10:100, 13:14, 50:60]
	"""
	return tuple(slice(s,s+1) if type(s) is int else s for s in slice_tuple)

def get_indices(
		a : np.ndarray, 
		slice_tuple : tuple[slice|int,...] | np.ndarray[int] | None,
		as_tuple : bool = True
	) -> np.ndarray | tuple[np.ndarray,...]:
	"""
	Returns the indices of an array slice. Indicies will select the sliced part of an array.
	
	a
		The array to get the indicies of a slice of
	slice_tuple
		A tuple of length `a.ndim` that specifies the slices
	as_tuple
		If True will return an `a.ndim` length tuple, otherwise will return a numpy array.
	"""
	
	if slice_tuple is None:
		slice_tuple = tuple([slice(None)]*a.ndim)
	
	_lgr.debug(f'BEFORE {slice_tuple=}')
	slice_tuple = unsqueeze(slice_tuple) # want a.ndim == a[sliced_idxs].ndim
	_lgr.debug(f'AFTER {slice_tuple=}')
	
	slice_idxs = np.indices(a.shape)[(slice(None),*slice_tuple)] 
	
	return tuple(slice_idxs) if as_tuple else slice_idxs 
	

def iter_indices(
		a : np.ndarray, 
		slice_tuple : tuple[slice|int,...] | np.ndarray[int] | None = None,
		group : tuple[int,...] = tuple(),
		squeeze=True
	) -> Iterator[tuple[np.ndarray]]:
	"""
	Iterator that returns the sliced indices of a sliced array.
	
	a
		The array to get the indicies of a slice of
	slice_tuple
		A tuple of length `a.ndim` that specifies the slices
	group
		Axes that are not iterated over (i.e. they are grouped together). E.g.
		if a.shape=(3,5,7,9) and group=(1,3), then on iteration the indices
		`idx` select a slice from `a` such that a[idx].shape=(5,9).
	squeeze
		Should non-grouped axes be merged (need one for loop for all of them),
		or should they remain separate (need a.ndim - len(group) loops).
	"""
	
	group = tuple(group)
	_lgr.debug(f'{a.shape=} {slice_tuple=} {group=} {squeeze=}')
	sliced_idxs = get_indices(a, slice_tuple, as_tuple=False)
	
	_lgr.debug(f'BEFORE {sliced_idxs.shape=}')
	if squeeze:
		sliced_idxs = nph.axes.merge(sliced_idxs, tuple(1+x for x in nph.axes.not_in(a,group)), 0)
	else:
		sliced_idxs = np.moveaxis(sliced_idxs, (1+x for x in group), (x for x in range(len(group),sliced_idxs.ndim)))
		sliced_idxs = np.moveaxis(sliced_idxs, 0, -(len(group)+1))
		
	_lgr.debug(f'AFTER {sliced_idxs.shape=}')
	return (tuple(x) for x in sliced_idxs)
	



"""
Operations on distributions derived from empirical data
"""
import dataclasses as dc
from numbers import Number
from typing import Any

import numpy as np

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'DEBUG')


class EmpiricalDistribution:
	"""
	Represents a probability distribution found through empirical data
	"""
	def __init__(self, data : np.ndarray):
		self._data = data
	
	def whole_cdf(self) -> tuple[np.ndarray, np.ndarray]:
		sorted_data = np.sort(self._data)
		return sorted_data, self.cdf(sorted_data)
	
	def cdf(self, value : Number | np.ndarray) -> Number | np.ndarray:
		"""
		Returns the cumulative density of `value` (i.e. fraction of datapoints less than `value`), can be passed an array.
		"""
		return np.interp(
			value,
			np.take_along_axis(self._data, np.argsort(self._data, axis=None), axis=None),
			np.linspace(0,1,self._data.size),
			left=0,
			right=1
		)
	
	def ppf(self, prob : Number | np.ndarray) -> Number | np.ndarray:
		"""
		Returns the value of the distribution at cumulative probability `prob`. I.e. `ppf(0.5)` is the median, can be passed an array
		"""
		return np.interp(
			prob,
			np.linspace(0,1,self._data.size),
			np.take_along_axis(self._data, np.argsort(self._data, axis=None), axis=None),
			left=np.nan, # cannot have probabilites > 1 or < 1, so return NAN if out of bounds.
			right=np.nan
		)

	def pdf(self, nbins=100) -> Number | np.ndarray:
		"""
		Returns the probability density function of the distribution
		"""
		n_bin_centers = nbins - 1
		x = np.linspace(np.min(self._data), np.max(self._data), n_bin_centers)
		dx = np.diff(x)
		bins = np.zeros((x.shape[0]+1,*x.shape[1:]))
		bins[0] = x[0] - dx[0]/2
		bins[1:-1] = x[:-1] + dx/2
		bins[-1] = x[-1] + dx[-1]/2
				
		counts = np.zeros((n_bin_centers,), dtype=float)
		for idx in range(0, bins.shape[0]-1):
			counts[idx] = np.count_nonzero((bins[idx] < self._data) & (self._data <= bins[idx+1]))
		
		return bins, counts/self._data.size
		
		
		
		





class Pointer:
	"""
	Lets the user set and retrieve referenced variables like a pointer does.
	Not sure exactly how to emulate pointers to normal variables yet.

	USAGE
	a_ptr = Pointer(7)
	b_ptr = Pointer(8)

	a_ptr.ref = b_ptr.ref
	print(a_ptr) # '8'

	b_ptr.val = 10
	print(a_ptr) # '10'
	"""
	__slots__=("__ref",)
	def __init__(self, val):
		self.ref = [val]
	@property
	def ref(self):
		""" Get the "reference" object, i.e. the internal list"""
		return(self.__ref)
	@ref.setter
	def ref(self, ref_list):
		""" Set the reference object, i.e. change the internal list"""
		self.__ref = ref_list
	@property
	def val(self):
		""" get the value stored in the reference, i.e. the single list element"""
		return(self.__ref[0])
	@val.setter
	def val(self, val):
		""" set the value stored in the referece, i.e. the single list element"""
		self.__ref[0] = val



"""
Contains classes that assist with plotting histograms.
"""

import sys
from time import sleep
from typing import Any
import dataclasses as dc


import numpy as np
import matplotlib as mpl
mpl.use('TKAgg')
import matplotlib.pyplot as plt

import aopp_deconv_tool.plot_helper as plot_helper
from aopp_deconv_tool.plot_helper.base import Base, AxisDataMapping


@dc.dataclass(repr=False, eq=False, slots=True)
class PlotSet:
	"""
	Defines a set of plots that can be animated
	"""
	fig : mpl.figure.Figure
	title : str = 'Plot Set'
	plots : list[Base] = dc.field(default_factory=list)
	blit : bool = True
	cadence : int = 1
	
	# Internal attributes
	n_frames : int = dc.field(default=0, init=False)
	blit_regions : Any = dc.field(default=None, init=False)
	blit_regions_data : Any = dc.field(default=None, init=False)
	title_hdl : Any = dc.field(default=None, init=False)
	title_blit_region : Any = dc.field(default=None, init=False)
	title_blit_region_data : Any = dc.field(default=None, init=False)
		
	def attach_datasource(self, datasource : Any):
		"""
		loop through plots, any with no attached datasource will use this one
		"""
		for p in self.plots:
			if p.is_datasource_attached():
				continue
			p.attach_datasource(datasource, p.datasource_data_getter)
		return self
		
	def show(self):
		self.title_hdl = self.fig.suptitle(self.title.format(self=self))
		self.n_frames = 0
		self.fig.canvas.draw()
		plt.show(block=False)
		
		#if self.blit:
			#plt.pause(0.4)
			#self.blit_regions = [self.fig.canvas.copy_from_bbox(ax.bbox) for ax in self.fig.axes]
			#self.blit_regions = [self.fig.canvas.copy_from_bbox(ax.get_tightbbox()) for ax in self.fig.axes]
	
	
	def get_ax_bb(self, ax):
		return ax.get_tightbbox().expanded(1.2,1.2)
	
	def update(self):
		
		if self.blit:
			# do one full draw then animate with blit
			if self.n_frames == 0:
				
				for ax in self.fig.axes:
					ax.drawn = False
				
				for plot in self.plots:
					plot.update()
				plt.pause(0.1)
					
				# set axes invisible that will change
				self.title_hdl.set_visible(False)
				for p in self.plots:
					p.ax.set_visible(True if p.static_frame else False)
				
					
				self.fig.canvas.draw()
				self.fig.canvas.flush_events()
				
				# set axes visible again
				self.title_hdl.set_visible(True)
				for p in self.plots:
					p.ax.set_visible(True)
					
				# need axes visible to get their extent
				self.title_blit_region = self.title_hdl.get_window_extent().expanded(1.2,1.2)
				self.blit_regions = tuple(self.get_ax_bb(ax) for ax in self.fig.axes)
				
				# set axes invisible that will change
				self.title_hdl.set_visible(False)
				for p in self.plots:
					p.ax.set_visible(True if p.static_frame else False)
				
				self.fig.canvas.draw()
				self.fig.canvas.flush_events()
						# need axes invisible to get the blank background
				self.title_blit_region_data = self.fig.canvas.copy_from_bbox(self.title_blit_region)
				self.blit_regions_data = tuple(self.fig.canvas.copy_from_bbox(r) for r in self.blit_regions)
						
				# set axes visible again
				self.title_hdl.set_visible(True)
				for p in self.plots:
					p.ax.set_visible(True)
				
				self.fig.canvas.draw()
				self.fig.canvas.flush_events()
				
				
				self.n_frames += 1
				return
				
			if self.n_frames % self.cadence == 0 :	
				# remove last frame plots
				self.fig.canvas.restore_region(self.title_blit_region_data)
				for ax, blit_region_data in zip(self.fig.axes, self.blit_regions_data):
					self.fig.canvas.restore_region(blit_region_data)
					ax.drawn=False
			
				# blit new frame plots
			self.title_hdl.set_text(self.title.format(self=self))
			
			if self.n_frames % self.cadence == 0 :
				self.fig.draw_artist(self.title_hdl)
				self.fig.canvas.blit(self.title_blit_region)
			
			for plot in self.plots:
				plot.update()
				if self.n_frames % self.cadence == 0 :
					self.fig.canvas.blit(self.get_ax_bb(plot.ax))
		else:
			self.title_hdl = self.fig.suptitle(self.title.format(self))
			for plot in self.plots:
				plot.update()
			
			if self.n_frames % self.cadence == 0 :
				self.fig.canvas.draw()
		
		if self.n_frames % self.cadence == 0 :	
			self.fig.canvas.flush_events()
		
	
		self.n_frames += 1


@dc.dataclass(repr=False, eq=False, slots=True)
class Histogram(Base):
	title : str = 'Histogram'
	datasource_name : str = 'histogram datasource'
	
	# Order corresponds to which axes they are associated with (0,1,2,..) -> (x,y,z,...)
	axis_data_mappings : tuple[AxisDataMapping] = dc.field(default_factory= lambda :(
		AxisDataMapping('value','bins',limit_getter=plot_helper.LimRememberExtremes()),
		AxisDataMapping('count','_hist',limit_getter=plot_helper.LimRememberExtremes()))
	)
	nbins : int = 100
	
	# Internal
	_hist : np.ndarray = None
	_bins : np.ndarray = None
	
	
	@property
	def bins(self):
		return self._bins
	
	@bins.setter
	def bins(self, value):
		self._bins[:-1] = value
		self._bins[-1] = self._bins[-2] + (self._bins[-2] - self.bins[-3])
	
	@property
	def hist(self):
		return self._hist
	
	@hist.setter
	def hist(self, value):
		self._hist[1:-1] = value
	
	def update_plot_data(self, data):
		self.hist[1:-1], self.bins = np.histogram(data, bins=self.nbins, **self.plt_kwargs)
		self.hdl.set_data(self.bins, self.hist)
	
	def on_attach_datasource(self):
		super(Histogram, self).on_attach_datasource()
		# initialise data holders, need to have zeros at either end due to how `self.ax.step` works
		self._hist = np.zeros((self.nbins+2,))
		self._bins = np.linspace(0,1,self.nbins+2,endpoint=True)

	def on_attach_ax(self):
		super(Histogram, self).on_attach_ax()
		self.hdl = self.ax.step([],[], label=self.datasource_name)[0]
	

@dc.dataclass(repr=False, eq=False, slots=True)
class Image(Base):
	title : str = 'Image'
	datasource_name : str = 'image datasource'
	
	# Order corresponds to which axes they are associated with (0,1,2,..) -> (x,y,z,...)
	axis_data_mappings : tuple[AxisDataMapping] = dc.field(default_factory= lambda :(
		AxisDataMapping('x',None), 
		AxisDataMapping('y',None), 
		AxisDataMapping('brightness', '_z_data'))
	)
	
	_z_data : np.ndarray = None
	
	def update_plot_data(self, data):
		
		self._z_data = data
		if self.hdl is None:
			#self.x_mesh, self.y_mesh = np.indices(self._z_data.shape) #np.arange(0,data.shape[0],1)
			#self.y_mesh = np.arange(0,data.shape[1],1)
			#self.hdl = self.ax.pcolormesh(self.x_mesh, self.y_mesh, self._z_data, label=self.datasource_name, shading='nearest', **self.plt_kwargs)
			kwargs = dict(origin='lower', interpolation='none')
			kwargs.update(self.plt_kwargs)
			self.hdl = self.ax.imshow(self._z_data, label=self.datasource_name, **kwargs)
			#self.ax.legend()
		else:
			self.hdl.set_data(self._z_data)
	
	def on_attach_datasource(self):
		super(Image, self).on_attach_datasource()

	def on_attach_ax(self):
		super(Image, self).on_attach_ax()


@dc.dataclass(repr=False, eq=False, slots=True)
class VerticalLine(Base):
	title : str = 'Vertical Line'
	datasource_name : str = 'vertical line datasource'
	
	# Order corresponds to which axes they are associated with (0,1,2,..) -> (x,y,z,...)
	axis_data_mappings : tuple[AxisDataMapping] = dc.field(default_factory= lambda :(AxisDataMapping(None,None),))
	
	
	# Internal
	_x_pos : float = None
	
	def update_plot_data(self, data):
		self._x_pos = data
		if self.hdl is not None:
			self.hdl.remove()
		self.hdl = self.ax.axvline(self._x_pos, label=self.datasource_name, **self.plt_kwargs)
	
	def on_attach_datasource(self):
		super(VerticalLine, self).on_attach_datasource()

	def on_attach_ax(self):
		super(VerticalLine, self).on_attach_ax()


@dc.dataclass(repr=False, eq=False, slots=True)
class HorizontalLine(Base):
	title : str = 'Horizontal Line'
	datasource_name : str = 'horizontal line datasource'
	
	# Order corresponds to which axes they are associated with (0,1,2,..) -> (x,y,z,...)
	axis_data_mappings : tuple[AxisDataMapping] = dc.field(default_factory= lambda :(AxisDataMapping(None,None),))
	
	
	# Internal
	_y_pos : float = None
	
	def update_plot_data(self, data):
		
		self._y_pos = data
		if self.hdl is not None:
			self.hdl.remove()
		self.hdl = self.ax.axhline(self._y_pos, label=self.datasource_name, **self.plt_kwargs)
	
	def on_attach_datasource(self):
		super(HorizontalLine, self).on_attach_datasource()

	def on_attach_ax(self):
		super(HorizontalLine, self).on_attach_ax()


@dc.dataclass(repr=False, eq=False, slots=True)
class IterativeLineGraph(Base):
	title : str = 'Iterative Line Graph'
	datasource_name : str = 'iterative line datasource'
	
	axis_data_mappings : tuple[AxisDataMapping] = dc.field(default_factory= lambda : (
			AxisDataMapping(
				'iteration',
				'_x',
				limit_getter=plot_helper.LimRememberExtremes(plot_helper.LimFixed(0))
			),
			AxisDataMapping(
				'value',
				'_y',
				limit_getter=plot_helper.LimRememberExtremes(plot_helper.LimFixed(0))
			)
		)
	)
	
	_x : list = dc.field(default_factory=list)
	_y : list = dc.field(default_factory=list)
	
	def update_plot_data(self, data):
		#print(f'{self.datasource_name} {self._y}')
		self._x.append(len(self._x))
		self._y.append(data)
		self.hdl.set_data(self._x,self._y)
	
	def on_attach_datasource(self):
		super(IterativeLineGraph, self).on_attach_datasource()
		self._x = []
		self._y = []

	def on_attach_ax(self):
		super(IterativeLineGraph, self).on_attach_ax()
		self.hdl = self.ax.plot(self._x, self._y, label=self.datasource_name, **self.plt_kwargs)[0]



"""
Helper functions to operate on FITS headers.
"""
import re

import numpy as np

import astropy as ap
from astropy.wcs import WCS

from aopp_deconv_tool.numpy_helper.axes import AxesOrdering

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'DEBUG')


class DictReader:
	"""
	Put a dictionary into a format that we can insert into a FITS header file.
	Make all uppercase, replace dots with spaces, turn everything into a string
	and hope it's not over 80 characters long. Will need to use HIERARCH keys
	to have long parameter names, see 
	https://fits.gsfc.nasa.gov/fits_standard.html

	Also we should combine dictionaries so that we have a flat structure, one
	way to do this is to concatenate the keys for member dictionaries with the
	parent dictionary, i.e:
		{"some.key" : {"child_key_1" : X1, "child_key_2": X2}} 
	becomes
		{"some.key.child_key_1" : X1, "some.key.child_key_2": X2}
	and is turned into a FITS header like
		HIERARCH SOME KEY CHILD_KEY_1 = "X1"
		HIERARCH SOME KEY CHILD_KEY_2 = "X2"

	If we need to we can split header entries into a "name" and "value" format,
	e.g. the above example would become:
		PKEY1   = some.key.child_key_1
		PVAL1   = "X1"
		PKEY2   = some.key.child_key_2
		PVAL1   = "X2"
	"""
	
	def __init__(self,
			adict: dict,
			mode : str = 'standard', # 'hierarch' | 'standard',
			prefix : str = None,
			pkey_count_start : int = 0
		):
		self.prefix = prefix
		self.adict = adict
		self.mode = mode
		self._key_count_start = pkey_count_start
		

	def __iter__(self):
		self.fits_dict = {}
		self.key_count = self._key_count_start
		if self.mode == 'hierarch':
			self._to_fits_hierarch_format(
				self.adict, 
				prefix=(' '.join(['HIERARCH',self.prefix])) if self.prefix is not None else 'HIERARCH'
			)
		elif self.mode == 'standard':
			self._to_fits_format(self.adict, prefix=self.prefix)
		else:
			raise RuntimeError(f"Unknown mode \"{self.mode}\" for creating FITS header cards from a dictionary, known modes are ('hierarch', 'standard').")
		return(iter(self.fits_dict.items()))

	@staticmethod
	def find_max_pkey_n(hdr, n_max=1000):
		pkey_fmt = 'PKEY{}'
		pkey_fmt_iter = ((i,) for i in range(n_max))
		pkey_max = get_key_fmt_max(hdr, pkey_fmt, pkey_fmt_iter)
		
		return int(pkey_max[4:]) if pkey_max is not None else 0
	
	@staticmethod
	def remove_pkeys(hdr, n_max=1000):
		pkey_fmt = 'PKEY{}'
		pkey_fmt_iter = ((i,) for i in range(n_max))
		remove_key_fmt(hdr, pkey_fmt, pkey_fmt_iter)

	def _to_fits_hierarch_format(self, bdict, prefix=None):
		for key, value in bdict.items():
			fits_key = ('' if prefix is None else ' ').join([('' if prefix is None else prefix), key.replace('.', ' ').upper()])
			if type(value) is not dict:
				self.fits_dict[fits_key] = str(value)
			else:
				self._to_fits_hierarch_format(value, fits_key)
		return

	def _to_fits_format(self, bdict, prefix=None):
		prefix_str = '' if prefix is None else prefix
		prefix_join_str = '' if prefix is None else '.'

		for k, v in bdict.items():
			_k = prefix_join_str.join([prefix_str, k])
			_v = str(v)
			if type(v) is not dict:
				self._add_kv_to_fits_dict(_k,_v)
			else:
				self._to_fits_format(v, _k)
		return

	def _add_kv_to_fits_dict(self, k, v):
		k_key = f'PKEY{self.key_count}'
		v_key = f'PVAL{self.key_count}'
		key_fmt_str = "{: <8}"
		val_fmt_str = "{}" # astropy can normalise the value string if needed
		self.fits_dict[key_fmt_str.format(k_key)] = val_fmt_str.format(k)
		self.fits_dict[key_fmt_str.format(v_key)] = val_fmt_str.format(v)
		self.key_count += 1
		return

def remove_key_fmt(hdr : ap.io.fits.Header, key_fmt, key_fmt_arg_iter):
	last_valid_key = None
	for items in key_fmt_arg_iter:
		key = key_fmt.format(*items)
		hdr.remove(key, ignore_missing=True, remove_all=True)
	

def get_key_fmt_max(hdr : ap.io.fits.Header, key_fmt, key_fmt_arg_iter) -> str:
	last_valid_key = None
	for items in key_fmt_arg_iter:
		key = key_fmt.format(*items)
		if key in hdr:
			last_valid_key = key
	return last_valid_key

def get_celestial_axes(hdr : ap.io.fits.Header, wcsaxes_label=''):
	placeholders = ('x','y')
	fits_celestial_codes = ['RA--', 'DEC-', 'xLON','xLAT', 'xyLN', 'xyLT']
	iraf_celestial_codes = ['axtype=xi', 'axtype=eta']

	celestial_idxs = []
	for i in AxesOrdering.range(hdr['NAXIS']):
		if any(fits_code==hdr.get(f'CTYPE{i.fits}{wcsaxes_label}', '')[sum(fits_code.count(p) for p in placeholders):len(fits_code)] 
						for fits_code in fits_celestial_codes) \
				or any(iraf_code in hdr.get(f'WAT{i.fits}_{"001" if wcsaxes_label=="" else wcsaxes_label}', '') 
						for iraf_code in iraf_celestial_codes) \
				:
			celestial_idxs.append(i.numpy)
	return(tuple(celestial_idxs))


def get_spectral_axes(hdr, wcsaxes_label=''):
	fits_spectral_codes = ['FREQ', 'ENER','WAVN','VRAD','WAVE','VOPT','ZOPT','AWAV','VELO','BETA']
	iraf_spectral_codes = ['axtype=wave']
	
	spectral_idxs = []
	for i in AxesOrdering.range(hdr['NAXIS']):
		if any(fits_code==hdr.get(f'CTYPE{i.fits}{wcsaxes_label}', '')[:len(fits_code)] for fits_code in fits_spectral_codes):
			spectral_idxs.append(i.numpy)
		elif any(iraf_code in hdr.get(f'WAT{i.fits}_{"001" if wcsaxes_label=="" else wcsaxes_label}', '') for iraf_code in iraf_spectral_codes):
			spectral_idxs.append(i.numpy)
	return(tuple(spectral_idxs))

def get_polarisation_axes(hdr, wcsaxes_label=''):
	raise NotImplementedError

def get_time_axes(hdr, wcsaxes_label=''):
	raise NotImplementedError


def get_world_coords_of_axis(hdr, ax_idx, wcsaxes_label='', squeeze=True):
	"""
	Gets the world coordiates of an axis
	"""
	ax_idxs = tuple((x if type(x)==AxesOrdering else AxesOrdering(x, hdr['NAXIS'], 'numpy')) for x in (ax_idx if (type(ax_idx) in (list,tuple)) else (ax_idx,)))

	_lgr.debug(f'{ax_idxs=}')

	wcs = WCS(hdr, key=' ' if wcsaxes_label=='' else wcsaxes_label.upper(), naxis=tuple(x.fits for x in ax_idxs))
	
	ss = tuple(slice(0,int(hdr[f'NAXIS{x.fits}'])) for x in ax_idxs)
	
	coord_array = np.mgrid[ss].reshape(len(ax_idxs),-1).T
	
	return(np.squeeze(wcs.all_pix2world(coord_array, 0)))

def is_CDi_j_present(header, wcsaxes_label=''):
	# matches "CDi_ja" where i,j are digits of indeterminate length, and a is an optional uppercase letter in wcsaxes_label
	cdi_j_pattern = re.compile(r'CD\d+_\d+'+wcsaxes_label)
	for k in header.keys():
		amatch = cdi_j_pattern.match(k)
		if amatch is not None:
			return(True)
	return(False)

def get_axes_ordering(hdr, axes, ordering='numpy', wcsaxes_label=''):
	wcsaxes = hdr.get(f'WCSACES{wcsaxes_label}', hdr['NAXIS'])
	return tuple(AxesOrdering(_x, wcsaxes, ordering) for _x in axes)

def set_axes_transform(hdr, axis=None, unit=None, reference_value=None, delta_value=None, n_values=None, reference_pixel=None):
	new_hdr_keys= {
		f'CD{axis}_{axis}' : delta_value,	# Turn meters into Angstrom
		f'CUNIT{axis}' : unit,				# Tell FITS the units
		f'CRVAL{axis}' : reference_value,	# Set the value of the reference pixel in the spectral direction in the FITS file (center of first bin)
		f'CRPIX{axis}' : reference_pixel,	# Tell FITS the index of the reference pixel in the spectral direction (first pixel, FITS is 1-index based)
		f'NAXIS{axis}' : n_values,			# Tell FITS the number of spectral planes
	}
	
	
	
	hdr.update({(k,v if v is not None else hdr[k]) for k,v in new_hdr_keys.items()}) # Update the old header with the new values (in memory, not on disk) so we can use it to write the altered file.



def get_iwc_matrix(hdr, wcsaxes_label=''):
	"""
	Get intermediate world coordinate matrix (CDi_j or PCi_j matrix with scaling applied)
	
	IRAF does things differently, see <https://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.46.2794&rep=rep1&type=pdf>
	"""
	naxis = hdr['NAXIS']
	wcsaxes = hdr.get(f'WCSAXES{wcsaxes_label}',naxis)
	iwc_mat = np.zeros((wcsaxes,wcsaxes))
	CD_flag = is_CDi_j_present(hdr, wcsaxes_label)
	for i in (AxesOrdering(_x,wcsaxes,'numpy') for _x in range(0,wcsaxes)):
		for j in (AxesOrdering(_x,wcsaxes,'numpy') for _x in range(0,wcsaxes)):
			#print(i, j)
			#print(i.numpy, j.numpy)
			#print(i.fits, j.fits)
			if CD_flag:
				iwc_mat[i.numpy,j.numpy] = hdr.get(f'CD{i.fits}_{j.fits}{wcsaxes_label}',0)
			else:
				default = 1 if i==j else 0
				iwc_mat[i.numpy,j.numpy] = hdr.get(f'PC{i.fits}_{j.fits}{wcsaxes_label}',default)*hdr.get('CDELT{i.fits}{wcsaxes_label}',1)
	return(iwc_mat)

def set_iwc_matrix(hdr, iwc_matrix, wcsaxes_label='', CD_format=None):
	# for now assume that iwc_matrix is all we are getting, set CDELTi to 1
	wcsaxes = hdr_get_wcsaxes(hdr, wcsaxes_label)
	CD_format = is_CDi_j_present(hdr, wcsaxes_label) if CD_format is None else CD_format
	hdr_mat_str_fmt = ('CD' if CD_format else 'PC')+'{i}_{j}{wcsaxes_label}'
	hdr_CDELTi_fmt = 'CDELT{i}{wcsaxes_label}'
	for i in (AxesOrdering(_x,wcsaxes,'numpy') for _x in range(0,wcsaxes)):
		hdr[hdr_CDELTi_fmt.format(i=i, wcsaxes_label=wcsaxes_label)] = 1.0
		for j in (AxesOrdering(_x,wcsaxes,'numpy') for _x in range(0,wcsaxes)):
			hdr[hdr_mat_str_fmt.format(i=i, j=j, wcsaxes_label=wcsaxes_label)] = iwc_matrix[i,j]
	return




























